/****************************************************************************
* Copyright (C)2002 Lucent Technologies                                     *
* (clarkson@research.bell-labs.com http://cm.bell-labs.com/who/clarkson)    *
*                                                                           *
* This program is free software; you can redistribute it and/or             *
* modify it under the terms of the GNU General Public License               *
* as published by the Free Software Foundation; either                      *
* version 2 of the License, or (at your option) any later version.          *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License for more details.                              *
*                                                                           *
* You should have received a copy of the GNU General Public License         *
* along with this program; if not, write to the Free Software               *
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA,                 *
* or visit http://www.gnu.org/licenses/gpl.html                             *
*****************************************************************************/


#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <stdlib.h>
#include <assert.h>
#include <float.h>
#include <stdarg.h>

#include "pool.h"
#include "dist.h"


char input_format[] = "%lf";

#define MAXDIM 33

coord	**point_data;
size_t	point_data_alloc;

size_t	num_points;
static size_t	dim;

size_t	ED_poolnum;

void panic(char *fmt, ...) {
	va_list args;

	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	fflush(stderr);
	va_end(args);

	abort();
}

void use_data(type_data *t) {
	num_points = t->num_P;
	point_data = t->P;
	dim = t->dim;
}


size_t get_data(FILE *IN){	/* returns number of points obtained */

	size_t i;
	char buf[10000];
	char *b;
	coord p[MAXDIM];
	char *sp = "\f\n\r\t\v ";

	num_points = 0;
	point_data_alloc = 3000;

	dim = MAXDIM+1;
	while (1) {
		if (!fgets(buf,sizeof(buf), IN)) return num_points;
		for (i=0, b=strtok(buf, sp); b && i<MAXDIM; b = strtok(0,sp))
			sscanf(b, input_format, &p[i++]);
		if (b && i==MAXDIM) panic("point dimension too high");
		if (dim==MAXDIM+1) {
			dim = i;
			point_data=(coord**)calloc(point_data_alloc, sizeof(coord*));
			ED_poolnum = pool_init(dim*sizeof(coord));

		}
		if (dim!=i)
			panic("bad input:\n %s\n i=%d dim=%d\n", buf, i, dim);

		if (num_points>=point_data_alloc) {
			point_data_alloc *= 2;
			point_data = (coord**)realloc(point_data,
						point_data_alloc*sizeof(coord*));
		}
		point_data[num_points] = (coord*)pool_new(ED_poolnum);
		memcpy(point_data[num_points++], p, dim*sizeof(coord));
	}
	return 0;
}






float this_dist(size_t i, size_t j) {

	size_t k;
	double sum=0, diff;

	assert(i<num_points && j<num_points);

	for (k = 0; k<dim; k++) {
		diff = point_data[i][k] - point_data[j][k];
		sum += diff*diff;
	}

	return (float)sqrt(sum);
}


float this_dist2(size_t i, size_t j) {

	size_t k;
	double sum=0, diff;

	assert(i<num_points && j<num_points);

	for (k = 0; k<dim; k++) {
		diff = point_data[i][k] - point_data[j][k];
		sum += diff*diff;
	}

	return (float)sum;
}



void free_data(void){	/* free any type-specific data */
	pool_purge(ED_poolnum);
	free(point_data);
}

void this_show_point(size_t i){}

size_t this_color(size_t x) {if (point_data[x][1] < 500000) return 0; else return 1;}
