/****************************************************************************
* Copyright (C)2002 Lucent Technologies                                     *
* (clarkson@research.bell-labs.com http://cm.bell-labs.com/who/clarkson)    *
*                                                                           *
* This program is free software; you can redistribute it and/or             *
* modify it under the terms of the GNU General Public License               *
* as published by the Free Software Foundation; either                      *
* version 2 of the License, or (at your option) any later version.          *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License for more details.                              *
*                                                                           *
* You should have received a copy of the GNU General Public License         *
* along with this program; if not, write to the Free Software               *
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA,                 *
* or visit http://www.gnu.org/licenses/gpl.html                             *
*****************************************************************************/


typedef struct { float old, nnew;} ddat;
typedef struct { ddat min, max; } ddats;

typedef struct sb {
	sb_distance dist;		/* distance function used in construction */
	size_t n;				/* number of sites */
	size_t *DL;				/* lists of sites, put into an array */
	float *dDL;				/* for M, dist of touching site to current NN;
								for sb, dist of touching site to farthest site it is
							    NN to */
	size_t *offset;			/* list for site j starts at offset[j] in DL*/

	size_t end_thresh;		/* in construction, sites > end_thresh only
								added to affL; in searching,
								sites > end_thresh don't have list searched */
	size_t (*col)(size_t);	/* color function, sometimes used;
								given site number, returns color of site */
	sb_distance dist2;		/* distance function sometimes used for speed */
	

	/* the following are only for sb */


	ddats *nn;
	
	/*	these are all length n work arrays;
		they could probably be replaced
		by hash tables				*/
	char *seen_for;		/* 1 if this site has been pending */		
	size_t *seens;		/* stack of sites that have been pending */
	size_t ns;			/* number of sites in seens */

	size_t *cur_touching;	/* pointer into DL for site i */
	float *dist2q;		/* distance of site i to current q */
	size_t *ins_num;	/* insertion time of site i */
	size_t *ins; 		/* site inserted at time i */

	struct heap *psh;			/* heap used to track farthest site to this one */
} sb;

void show_sb(sb *);

void this_show_point(size_t);

void free_sb_i(sbp, int sbt);

#define make_storage(x, type, num)   {x = ( type *)calloc(( num )  , sizeof ( type ));    assert(x);}


typedef struct alist_str *alist;

typedef struct alist_str {
    size_t site;
    float fnn; /* fnn is max dist to site owned by touched site, just after this->site was added */
    float dDL;
    ddats nn;  /* of sites with this as old nn and *site* as new nn, ... */
    alist    next;
} alist_str;


void reset_seens(sbp sbt);

sb *init_sb(sb_distance dist, size_t n, float end_frac,
			char sb_flag, sb_distance dist2);



			/* instrumentation for stats; here nulled
				out to allow compilation without do_test.h */
#define stat_inc_change_nn ;
#define stat_inc_search_loop ;
#define stat_update_max_heap_entries(dum) ;


