/****************************************************************************
* Copyright (C)2002 Lucent Technologies                                     *
* (clarkson@research.bell-labs.com http://cm.bell-labs.com/who/clarkson)    *
*                                                                           *
* This program is free software; you can redistribute it and/or             *
* modify it under the terms of the GNU General Public License               *
* as published by the Free Software Foundation; either                      *
* version 2 of the License, or (at your option) any later version.          *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License for more details.                              *
*                                                                           *
* You should have received a copy of the GNU General Public License         *
* along with this program; if not, write to the Free Software               *
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA,                 *
* or visit http://www.gnu.org/licenses/gpl.html                             *
*****************************************************************************/


#include <stdio.h>
#include <math.h>
#include <time.h>
#include <stdlib.h>
#include <assert.h>
#include <float.h>
#include <stdarg.h>
#include <float.h>

#include "my_rand.h"
#include "gen_dist.h"
#include "pool.h"
#include "sb.h"
#include "MMM.h"
#include "Msbi.h"
#include "do_test.h"

size_t num_better, num_non_zeros;
size_t exact_sb = 0;

float err_ratio;

float dcounts[10] = {0};

sbtesttypes sbtesttype = test_closest;

char longprint = 0;

void print_prep(sbp sbt)
{
	printf(longprint ? 
		"prep. distance evals per site=%.1f\nD_list entries per site=%.1f\n"
		: "prep_evals=%.1f\nentries_per_site=%.1f\n"
		,dcounts[prep]/sbt->n
		,(float)size_lists(sbt)/sbt->n
	);
}

size_t	*shufmat;

void shuffle_up(size_t *mat, size_t n) {
	size_t i, j, t;

	for (i=0;i<n;i++){
		t = mat[i];
		mat[i] = mat[j = i + (size_t)((n-i)*my_rand())];
		mat[j] = t;
	}
}

float my_dist(size_t i, size_t j)
{
	dcounts[activity]++;
	return this_dist(shufmat[i], shufmat[j]);
}

float my_dist2(size_t i, size_t j) {return this_dist2(shufmat[i], shufmat[j]);}

void my_show_point(size_t i) {this_show_point(shufmat[i]);}

void make_shuffle(size_t num_all, size_t num_sites, size_t num_tot) {

	size_t i;

	assert(num_sites <= num_tot);
	assert(num_tot <= num_all);

	shufmat = (size_t*)malloc(num_all*sizeof(size_t));

	for (i=0; i<num_all; i++) shufmat[i] = i;


	/* shuffle_up(shufmat, num_all); */  /* for a piece of /usr/dict/words */

			/* usual case */

	shuffle_up(shufmat, num_sites);			
	shuffle_up(shufmat+num_sites, num_tot-num_sites);

}


void show_point(size_t i) {this_show_point(shufmat[i]);}

size_t closest(sb_distance dist, size_t num_sites, size_t q, float check_dist, char p) {

	size_t c=0, j;
	float cdist = FLT_MAX, jdist;
	activity = ignore; /* for managing distance eval stats */

	for (j=0; j<num_sites; j++) {
		jdist = dist(q,j);
		if (jdist*(1+1e-7) < check_dist) {
			if (p) printf("(%d %g) ", j, jdist);
			num_better++;
		}
		if (jdist >= cdist) continue;
		cdist = jdist;
		c = j;
	}
	if (p) printf("\n");
	return c;
}



size_t	exact_count=0;
static float k_nearest_dist;

static void get_kth_nearest_dist(SD* x)
{
	if (k_nearest_dist < x->key)
		k_nearest_dist = x->key;
}


size_t smin(size_t a, size_t b) {return (a<b) ? a : b;}

double stat_ans_size;

void test_sbk_q(sbp sbt, size_t num_sites, size_t q, float alpha, search_sb_f search_sb, search_M_f search_M)
{
	size_t k=5;
	stat_ans_size = k;

	k_nearest_dist = 0;
	stat_search_loop = stat_max_heap_entries = 0;
	activity = query;
	search_sb_k_nearest(sbt, q, k, &get_kth_nearest_dist);
	activity = ignore;

	num_better = 0;
	closest(my_dist, num_sites, q, k_nearest_dist, 0);
	assert(num_better==k-1);

	return;
}

static void noop_visit(SD* x) {}

void test_sbr_q(sbp sbt, size_t num_sites, size_t q, float alpha, search_sb_f search_sb, search_M_f search_M)
{
	size_t i, ans_size, true_ans_size;
	double r = 1;

	activity = ignore;
	ans_size=0;
	while ((ans_size = search_sb_fixed_r(sbt, q, (float)r, &noop_visit)) >= smin(sbt->n, 5)) {r /= 1.2;}
	while ((ans_size = search_sb_fixed_r(sbt, q, (float)r, &noop_visit)) == 0) {r *= 1.2;}

	stat_search_loop = stat_max_heap_entries = 0;
	activity = query;
	ans_size = search_sb_fixed_r(sbt, q, (float)r, &noop_visit);
	activity = ignore;

	true_ans_size = 0;
	for (i=0; i<sbt->n; i++) {
		if (my_dist(q,i) < r)
			true_ans_size++;
	}

	assert(ans_size==true_ans_size);

	stat_ans_size = ans_size;

	return;
}






static void null_visit(SD *x) {}

void test_sbt_q(sbp sbt, size_t num_sites, size_t q, float alpha,
				search_sb_f search_sb, search_M_f search_M) {

	size_t	exact_closest,
		sb_closest;
	float	ec_dist,
		sb_dist;

	activity = query;
	sb_closest = search_sb ? search_sb(sbt, q, alpha) : search_M(sbt, q);
	activity = ignore;
	sb_dist = my_dist(sb_closest, q);
	activity = a_verify;
	search_sb_fixed_r(sbt, q, sb_dist, &null_visit);

	if (exact_sb && alpha==1)
		return;

	exact_closest = closest(my_dist, num_sites, q, sb_dist, 0); 
	ec_dist = my_dist(exact_closest, q);
	if (ec_dist>sb_dist*(1+1e-6))
		printf("sb_dist=%G, ec_dist=%G\n", sb_dist, ec_dist);
	if (ec_dist>=sb_dist)
		exact_count++;
	else if (0) {
		printf("\nq num=%d, q=", q);show_point(q);
		printf("\nsb_closest %d, distance %G:", sb_closest, sb_dist);
		show_point(sb_closest);
		printf("\nexact_closest %d, distance %G:", exact_closest, ec_dist);
		show_point(exact_closest); printf("\n");fflush(stdout);
	}
	if (ec_dist!=0.0) {
		err_ratio += sb_dist/ec_dist;
		num_non_zeros++;
	}
}


void test_sbt(sbp sbt, size_t num_sites, size_t start_tests,
	size_t num_tests, float alpha, search_sb_f search_sb, search_M_f search_M) {

	size_t j;
	double s_stat_max_heap_entries = 0, s_stat_search_loop = 0, s_stat_ans_size = 0;
	char short_fmt[] =
		"%.1f\t%.2f\t%3.2f\t%.2f\t%.1f\t%.1f\t%.0f\n";
	char long_fmt[] = 
		"ave query distance evals=%.1f\n"
		"ave error ratio = %.1f\n"
		"ave queries answered exactly = %.1f\n"
		"ave num sites closer than answer = %.1f\n"
		"ave max heap size =%.1f\n"
		"ave iterations of query loop = %.1f\n"
		"%% of ave verification distance evals to query=%.0f\n";


	dcounts[query] = dcounts[a_verify] = err_ratio = 0;
	exact_count = num_better = num_non_zeros = 0;

	for (j=0; j<num_tests; j++) {
		if (0 && j%100==0) {fprintf(stderr, "*");fflush(stderr);}
		stat_search_loop = stat_max_heap_entries = 0;
		switch(sbtesttype) {
			case test_fixed_radius:	test_sbr_q(sbt, num_sites, j+start_tests, alpha, search_sb, search_M); break;
			case test_k_nn:			test_sbk_q(sbt, num_sites, j+start_tests, alpha, search_sb, search_M); break;
			case test_closest:		test_sbt_q(sbt, num_sites, j+start_tests, alpha, search_sb, search_M); break;
		}
		s_stat_max_heap_entries += stat_max_heap_entries;
		s_stat_search_loop += stat_search_loop;
		s_stat_ans_size += stat_ans_size;
	}

	if (sbtesttype!=test_closest) {
		printf(longprint
			? 
				"ave query distance evals=%.1f\n"
				"ave max heap size =%.1f\n"
				"ave iterations of query loop = %.1f\n"
				"ave answer size = %.1f\n"
			:
				"%.1f\t%.2f\t%3.2f\t%.1f\n"
			,dcounts[query]/num_tests
			,s_stat_max_heap_entries/num_tests
			,s_stat_search_loop/num_tests
			,s_stat_ans_size/num_tests
		);
		return;
	}

	if (exact_sb && alpha==1 && search_sb) {
		err_ratio = (float)num_tests;
		num_non_zeros = num_tests;
	}

	printf(longprint ? long_fmt : short_fmt
		,dcounts[query]/num_tests
		,(err_ratio/num_non_zeros-1)*100
		,(float)exact_count/num_tests*100
		,(float)num_better/num_tests
		,s_stat_max_heap_entries/num_tests
		,s_stat_search_loop/num_tests
		,dcounts[a_verify]/dcounts[query]*100
	 );
}
